#include <SDL.h>
#include <math.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>

#define XSIZE 512
#define YSIZE 512

int mousex,mousey;

SDL_Surface *thescreen;
void scrlock(void)
{
	if ( SDL_LockSurface(thescreen) < 0 )
	{
		fprintf(stderr, "Couldn't lock display surface: %s\n",
							SDL_GetError());
		return;
	}
}
void scrunlock(void)
{
	SDL_UnlockSurface(thescreen);
}

void update(void)
{
	SDL_UpdateRect(thescreen, 0, 0, 0, 0);
}

unsigned long maprgb(int r,int g,int b)
{
	return SDL_MapRGB(thescreen->format,r,g,b);
}

void clear(void)
{
int i;
	for(i=0;i<YSIZE;++i)
		memset(thescreen->pixels + i*thescreen->pitch, 0, XSIZE*4);
}
void unmaprgb(Uint32 x,Uint32 y, Uint8 *r, Uint8 *g, Uint8 *b)
{
	if(x<XSIZE && y<YSIZE)
		SDL_GetRGB(*(Uint32 *)(thescreen->pixels+y*thescreen->pitch+(x<<2)),
			thescreen->format, r,g,b);
}

void colordot(Uint32 x,Uint32 y,int c)
{
	if(x<XSIZE && y<YSIZE)
		*(Uint32 *)(thescreen->pixels+y*thescreen->pitch+(x<<2))=c;
}
void colorrect(int x, int y, int w, int h, int c)
{
int i,j;
	for(j=0;j<h;++j)
		for(i=0;i<w;++i)
			colordot(x+i, y+j, c);
}
void circle(int cx,int cy,int radius,int c)
{
int x,y,e;

	x=0;
	y=radius;
	e=3-(radius<<1);
	while(x<=y)
	{
		colordot(cx+x,cy+y,c);
		colordot(cx-x,cy+y,c);
		colordot(cx+x,cy-y,c);
		colordot(cx-x,cy-y,c);
		colordot(cx+y,cy+x,c);
		colordot(cx-y,cy+x,c);
		colordot(cx+y,cy-x,c);
		colordot(cx-y,cy-x,c);
		if(e<0)
			e+=(x<<2)+6;
		else
		{
			e+=((x-y)<<2)+10;
			--y;
		}
		++x;
	}
}


#define MAX_CIRCLES 50000

struct circle {
	int used;
	double x;
	double y;
	double r;
	int red, green, blue;
} circles[MAX_CIRCLES];
int numcircles;

inline double min(double v1, double v2)
{
	return v1<v2 ? v1 : v2;
}



void makecircles(int n)
{
int i,j;
double x, y, r, r2, dx, dy;
struct circle *c;
int bad;
	for(i=0;i<n;++i)
	{
		if(circles[i].used) continue;
		for(bad = 1;bad;)
		{

			x = rand()%XSIZE;
			y = rand()%YSIZE;
			r = x;
			r = min(r, XSIZE-1-x);
			r = min(r, y);
			r = min(r, YSIZE-1-y);

			for(j=0;j<n;++j)
			{
				c = circles+j;
				if(!c->used) continue;

				dx = c->x - x;
				dy = c->y - y;
				r2 = sqrt(dx*dx + dy*dy);
				if(r2 <= c->r) break; // inside this circle
				r = min(r, r2 - c->r);
			}
			if(r == 0.0) continue;
			dx = x - XSIZE/2;
			dy = y - YSIZE/2;
			if(dx*dx + dy*dy <= r*r) continue;
			if(j<n) continue;
			bad = 0;
		}
		c = circles+i;
		c->x = x;
		c->y = y;
		c->r = r;
		c->used = 1;
		c->red = rand()&255;
		c->green = rand()&255;
		c->blue = rand()&255;
	}
	numcircles = n;
}

void saveppm(char *fn)
{
unsigned char temp[XSIZE*3];
int f;
int x,y, x3;

	f = open(fn, O_WRONLY|O_CREAT|O_TRUNC, 0644);
	sprintf((char *)temp, "P6\n%d %d\n255\n", XSIZE, YSIZE);
	write(f, (char *)temp, strlen((char *)temp));

	for(y=0;y<YSIZE;++y)
	{
		for(x=0;x<XSIZE;++x)
		{
			x3 = x*3;
			unmaprgb(x, y, temp+x3, temp+x3+1, temp+x3+2);
/*
			if(gd(x, y))
			{
				temp[x3] = 255;
				temp[x3+1] = 128 * x / XSIZE;
				temp[x3+2] = 0;
			} else
				temp[x3] = temp[x3+1] = temp[x3+1] = 0;
*/
		}
		write(f, temp, sizeof(temp));
	}
	close(f);
}


void drawsphere(int r,int bx,int by,int vx,int vy,int vz,int rt,int gt,int bt)
{
float x,y,y2,z,radius;
int i,j,j2,k,rint;
float tvx,tvy,tvz;
float dot;
unsigned char red[256],green[256],blue[256];
float lvx,lvy,lvz;
int rr,gg,bb;
int row,col;
int ty;
int left, right;

	for(i=0;i<256;++i)
	{
		if(i<232)
		{
			red[i]=224;
			green[i]=224;
			blue[i]=224;
		}
		else if(i<250)
		{
			red[i]=rt;
			green[i]=gt;
			blue[i]=bt;
		} else
			red[i]=green[i]=blue[i]=0;
	}

	lvx=-1.0;
	lvy=1.0;
	lvz=1.5;
	radius=sqrt(lvx*lvx+lvy*lvy+lvz*lvz);
	lvx/=radius;
	lvy/=radius;
	lvz/=radius;

	tvx=vx;
	tvy=vy;
	tvz=vz;
	radius=sqrt(tvx*tvx+tvy*tvy+tvz*tvz)/127.5;
	tvx/=radius;
	tvy/=radius;
	tvz/=radius;

	radius=r;
	rint=r;
	row=by;
	col=bx-rint;
	ty = row - rint;
	if(ty<0) ty = 0;
	for(j=ty - by;j<=rint && ty<YSIZE;++j, ++ty)
	{
		int tx;
		y=-j/radius;
		y2=y*y;
		j2=j*j;
		left = sqrt(radius*radius - (double)j*j);
		right = bx + left;
		left = bx - left;
		if(right<0) continue;
		if(left>=XSIZE) continue;

		tx = left;
		if(tx<0) tx = 0;
		if(right>XSIZE-1) right=XSIZE-1;
		for(i=tx - bx;tx < right;++i,++tx)
		{
			x=i/radius;
			z=1.0 - x*x - y2;
			if(z<0.0) continue;
			z=sqrt(z);
			k=127.5+x*tvx+y*tvy+z*tvz;
			dot=(x*lvx+y*lvy+z*lvz+1)/2;
			dot*=dot*dot;
			rr=red[k];
			gg=green[k];
			bb=blue[k];

			if(dot>=0.98)
			{
				rr=(rr+3*255)>>2;
				gg=(gg+3*255)>>2;
				bb=(bb+3*255)>>2;
			}

			rr*=dot;
			gg*=dot;
			bb*=dot;

			colordot(tx, ty, maprgb(rr,gg,bb));
		}
	}
}


void iterate(void)
{
int i;
struct circle *c;
double dx, dy, comp;

	dx = XSIZE > YSIZE ? XSIZE : YSIZE;
	dx /= 2.0;
	comp = sqrt(2.0 * dx * dx);

	makecircles(1000);
	scrlock();
	clear();

	for(i=0;i<numcircles;++i)
	{
		c=circles + i;
//		circle(c->x, c->y, c->r, ~0);
		drawsphere(c->r, c->x, c->y, XSIZE/2 - c->x, c->y - YSIZE/2, 500,
			c->red, c->green, c->blue);
	}
	scrunlock();
	update();
	for(i=0;i<numcircles;++i)
	{
		c = circles+i;
		if(!c->used) continue;
#define ZOOM 1.02
		c->x = XSIZE/2 + ZOOM*(c->x - XSIZE/2);
		c->y = YSIZE/2 + ZOOM*(c->y - YSIZE/2);
		c->r *= ZOOM;

		dx = c->x - XSIZE/2;
		dy = c->y - YSIZE/2;
		if(sqrt(dx*dx + dy*dy) - c->r >= comp)
			c->used = 0;

	}

}

int main(int argc,char **argv)
{
int code;
SDL_Event event;
unsigned long videoflags;
int done=0;
	srand(time(0));
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 )
	{
		fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
		exit(1);
	}
	videoflags = 0;
	thescreen = SDL_SetVideoMode(XSIZE, YSIZE, 32, videoflags);
	if ( thescreen == NULL )
	{
		fprintf(stderr, "Couldn't set display mode: %s\n",
							SDL_GetError());
		exit(5);
	}

	while(!done)
	{
		iterate();
//		SDL_Delay(50);
		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
			case SDL_MOUSEMOTION:
				mousex=event.motion.x;
				mousey=event.motion.y;
				break;
			case SDL_MOUSEBUTTONDOWN:
				mousex=event.button.x;
				mousey=event.button.y;
				break;
			case SDL_KEYDOWN:
				code=event.key.keysym.sym;
				if(code==SDLK_ESCAPE) done=1;
				break;
			case SDL_QUIT:
				done=1;
				break;
			}
		}
	}
//	saveppm("/ram/bigpic");
	SDL_Quit();
	return 0;
}
